
# Author : damien KROTKINE (dams@gentoo.org)
#
#
# Copyright (C) 2005 damien KROTKINE (dams@gentoo.org)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; only version 2, not any other version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

package Libconf::Templates::Lilo;
use strict;
use vars qw(@ISA);
use Libconf qw(:helpers);
use Libconf::Templates;
use Libconf::Templates::Generic::KeyValue;
use Libconf::Templates::Generic::Value;
@ISA = qw(Libconf::Templates::Generic::KeyValue);

=head1 NAME

Libconf::Templates::Lilo

=head1 DESCRIPTION

Libconf::Templates::Lilo - Libconf low level template to handle lilo configuration files

=head1 SYNOPSIS

 my $template = new Libconf::Templates::Lilo({ filename => '/etc/lilo.conf' });
 $template->edit_atom(-1, {
                          key => 'label',
                          value => 'linux',
                          sections => [ {name => '/boot/bzImage'} ],
                         },
                          {
                          key => 'label',
                          value => 'linux_foo',
                          sections => [ {name => '/boot/bzImage_foo'} ],
                         });
 ...
 (see L<Libconf::Templates> for transformation methods on $template)
 ...
$template->write_conf();

=head1 CONSTRUCTOR

=over

=item B<new($options)>

creates the template

 my $template = new Libconf::Templates::Lilo({ filename => '/etc/lilo.conf' });

B<options>

filename [B<type> : STRING, B<default> : ''] : the filename of the config file
you want to work on. Can be read and written lately by using set_filename and
get_filename.

Others options taken from L<Libconf::Templates::Generic::KeyValue> and
L<Libconf::Templates::Generic::Value> can be also set here.

=back

=head1 GENERAL METHODS

See L<Libconf::Templates> for the general list of methods you can call on this
template.

=head1 SPECIFIC METHODS

There is no specific methods

=cut

# my @global_v_sections = qw(change-rules);
# my @global_kv_sections = qw(disk);
# my @global_v_options = qw(bmp-retain compact el-torito-bootable-CD fix-table
# geometric ignore-table large-memory lba32 linear lock mandatory nowarn optional
# prompt restricted single-key static-BIOS-codes suppress-boot-time-BIOS-data
# read-only read-write);
# my @global_kv_options = qw(backup bios-passes-dl bitmap bmp-colors bmp-table
# bmp-timer boot default delay disktab force-backup install map menu-title
# menu-scheme message password raid-extra-boot serial timeout verbose vmdefault
# append ramdisk root vga);
# my @kernel_v_options = qw(read-only read-write);
# my @kernel_kv_options = qw(addappend append initrd literal ramdisk root vga);
# my @alternate_kv_options = qw(loader table table boot-as map-drive);
# my @alternate_v_options = qw(master-boot unsafe);
# my @alternate_v_sections = qw(change);
# my @common_v_options = qw(bmp-retain lock optional mandatory restricted bypass vmwarn vmdisable);
# my @common_kv_options = qw(label alias fallback password);

sub new {
    my ($class, $args) = @_;
    my $value_template = new Libconf::Templates::Generic::Value(init_args($args, {
                                                                                  regexp => '^\s*(\S+)\s*$',
                                                                                  output_function => sub { my ($atom) = @_; ($atom->{value}, scalar(@{$atom->{sections}})) },
                                                                                 }));
    my $self = bless $class->SUPER::new(init_args($args, {
                                                    filename => '',
                                                    value_template => $value_template,
                                                    separator_char => '=',
                                                    allow_space => 0,
                                                    handle_quote => 0,
                                                    handle_multiples_lines => 0,
                                                    handle_backslash => 1,
                                                    accept_empty_value => 1,
                                                    template_name => 'Lilo',
                                                   })), $class;
    $value_template->{atoms} = $self->{atoms};
    $self;
}

sub _parse_chunk {
    my ($self, $in, $atom, $parser_env) = @_;

    # normal sections
    my $section_regexp = '^\s*(image|other)=(\S+)\s*$';
    if ($in =~ s/$section_regexp//) {
        $atom->{type} = 'SECTION';
        $atom->{type2} = $1;
        $atom->{section_name} = $2;
        # in this case, we end the previous section
        pop @{$parser_env->{current_sections}};
        $atom->{sections} = [ @{$parser_env->{current_sections}} ];
        push @{$parser_env->{current_sections}}, { name => $2 };
        return ($in, 'stop');
    }

    # key=value rule
    my ($out, $command) = $self->SUPER::_parse_chunk($in, $atom);
    $command eq 'failed' or return ($out, $command);

    # single options rule (value)
    $self->{value_template}->_parse_chunk($in, $atom, $parser_env);

}

sub _output_atom {
    my ($self, $atom, $parser_env) = @_;
    if ($atom->{type} eq 'SECTION') {
        # sections output
        return ($atom->{type2} . '=' . $atom->{section_name}, scalar(@{$atom->{sections}}));
    } elsif ($atom->{type} eq 'VALUE') {
        # value output
        return $self->{value_template}->_output_atom($atom, $parser_env);
    }
    # keyvalue output;
    $self->SUPER::_output_atom($atom, $parser_env);
}
 
1;
